import functools

from typing import Union, Callable

from ._ldaps import _get_session


def initialize(*, address: str, port: Union[int, str], domain: str, username: str, password: str) -> Callable:
    """ Initialize an LDAP session and get a connector context manager.

    :param address: The hostname or IP address of the domain controller to connect to.
    :param port: The listening LDAPS port (usually 636).
    :param domain: The domain FQDN.
    :param logon_name: The username to bind with.
    :param password: The password to bind with.

    :raises exceptions.UserAccountLocked: if the supplied user account is locked.
    :raises exceptions.PasswordExpired: if the supplied user account has an expired password.
    :raises exceptions.InvalidCredentials: if authentication using the supplied credentials failed.
    :raises exceptions.ServerDown: if connectivity to the supplied address failed.

    :returns: A context manager that yields an opened LDAPS connection.

    The returned context manager should only be called using a `with statement
    <https://docs.python.org/3.8/reference/compound_stmts.html#with>`_.

    All user/group operations must happen inside the `with` block. You do not need to know how to
    open or close an LDAPS connection - the context manager handles all of this for you.

    You can re-use an initialized session as often as necessary. This makes it easier to separate
    offline programming logic from Active Directory module operations as much as possible.

    See the example below, or the more complete :ref:`active_directory:Example Script` to view it in
    action.

    .. code-block:: python

        session = client.initialize(
            address='host',
            port='636',
            domain='my.domain.com',
            username='username',
            password='password'
        )

        with session() as connection:
            user = users.search(connection, sAMAccountName='my.user')

    """
    return functools.partial(_get_session, address, port, domain, username, password)
