import logging

from datacaster.classes import CastObject, CastDataClass

from typing import TypedDict, Optional, List

logger = logging.getLogger(__name__)


def _format_date_string(value):
    return value if not value else value.date_object.strftime("%H:%M:%S %d/%m/%Y")


def _get_decoded_ldap_string(value):
    return value.string if hasattr(value, "string") else None


class DomainController(TypedDict):
    """
- Contains Domain Controller connection details

- It is used as an _input_ to many of the actions in this plugin

- The example below shows the dictionary structure in YAML

- See the ___required___ and _optional_ keys underneath the example

!!! warning "Supplying Credentials"

    You should ___always___ use a PPA Vault integration to provide credentials to a plugin action.

!!! example

    ```yaml linenums="1"
    domain_controller:
      address: 1.2.3.4
      domain: example.domain.net
      port: 636
      username: [username]
      password: [password]
    ```

!!! danger "Required Keys"

    __address__: Domain Controller IP or DNS address.

    __domain__: FQDN of the Active Directory domain.

    __username__: Username for authentication.

    __password__: Password for authentication.

!!! info "Optional Keys"

    __port__: The LDAPS port on the Domain Controller (defaults to `636`).

    """

    address: str
    port: Optional[int]
    domain: str
    username: str
    password: str


class PasswordAge(TypedDict):
    days: Optional[int]
    hours: Optional[int]
    minutes: Optional[int]


class User(TypedDict):
    """
- Contains Active Directory User LDAP attributes

- Expand the sections below for more information

??? info "Active Directory User Keys"

    __accountExpires__: Date string or null [[Account-Expires](https://docs.microsoft.com/en-us/windows/win32/adschema/a-accountexpires)].

    __badPasswordTime__: Date string or null [[Bad-Password-Time](https://docs.microsoft.com/en-us/windows/win32/adschema/a-badpasswordtime)].

    __badPwdCount__: Integer [[Bad-Pwd-Count](https://docs.microsoft.com/en-us/windows/win32/adschema/a-badpwdcount)].

    __carLicense__: String or null [[carLicense](https://docs.microsoft.com/en-us/windows/win32/adschema/a-carlicense)].

    __cn__: String or null [[Common-Name](https://docs.microsoft.com/en-us/windows/win32/adschema/a-cn)].

    __countryCode__: String or null [[Country-Code](https://docs.microsoft.com/en-us/windows/win32/adschema/a-countrycode)].

    __displayName__: String or null [[Display-Name](https://docs.microsoft.com/en-us/windows/win32/adschema/a-displayname)].

    __distinguishedName__: String [[Obj-Dist-Name](https://docs.microsoft.com/en-us/windows/win32/adschema/a-distinguishedname)].

    __gidNumber__: Integer or null [[gidNumber](https://docs.microsoft.com/en-us/windows/win32/adschema/a-gidNumber)].

    __givenName__: String or null [[Given-Name](https://docs.microsoft.com/en-us/windows/win32/adschema/a-givenname)].

    __info__: String or null [[Comment](https://docs.microsoft.com/en-us/windows/win32/adschema/a-info)].

    __lastLogoff__: Date string or null [[Last-Logoff](https://docs.microsoft.com/en-us/windows/win32/adschema/a-lastlogoff)].

    __lastLogon__: Date string or null [[Last-Logon](https://docs.microsoft.com/en-us/windows/win32/adschema/a-lastlogon)].

    __lastLogonTimestamp__: Date string or null [[Last-Logon-Timestamp](https://docs.microsoft.com/en-us/windows/win32/adschema/a-lastlogontimestamp)].

    __lockoutTime__: Date string or null [[Lockout-Time](https://docs.microsoft.com/en-us/windows/win32/adschema/a-lockouttime)].

    __loginShell__: String or null [[loginShell](https://docs.microsoft.com/en-us/windows/win32/adschema/a-loginShell)].

    __logonCount__: Integer [[Logon-Count](https://docs.microsoft.com/en-us/windows/win32/adschema/a-logoncount)].

    __mail__: String or null [[E-mail-Addresses](https://docs.microsoft.com/en-us/windows/win32/adschema/a-mail)].

    __manager__: String or null [[Manager](https://docs.microsoft.com/en-us/windows/win32/adschema/a-manager)].

    __memberOf__: List of [Group Dictionaries](#group) [[Is-Member-Of-DL](https://docs.microsoft.com/en-us/windows/win32/adschema/a-memberof)].

    __name__: String or null [[RDN](https://docs.microsoft.com/en-us/windows/win32/adschema/a-name)].

    __objectGUID__: String [[Object-Guid](https://docs.microsoft.com/en-us/windows/win32/adschema/a-objectguid)].

    __objectSid__: String [[Object-Sid](https://docs.microsoft.com/en-us/windows/win32/adschema/a-objectsid)].

    __physicalDeliveryOfficeName__: String or null [[Physical-Delivery-Office-Name](https://docs.microsoft.com/en-gb/windows/win32/adschema/a-physicaldeliveryofficename)].

    __pwdLastSet__: Date string or null [[Pwd-Last-Set](https://docs.microsoft.com/en-us/windows/win32/adschema/a-pwdlastset)].

    __sAMAccountName__: String [[Sam-Account-Name](https://docs.microsoft.com/en-us/windows/win32/adschema/a-samaccountname)].

    __sAMAccountType__: Integer [[Sam-Account-Type](https://docs.microsoft.com/en-us/windows/win32/adschema/a-samaccounttype)].

    __sn__: String or null [[Surname](https://docs.microsoft.com/en-us/windows/win32/adschema/a-sn)].

    __telephoneNumber__: String or null [[Telephone-Number](https://docs.microsoft.com/en-us/windows/win32/adschema/a-telephonenumber)].

    __thumbnailPhoto__: String or null [[Picture](https://docs.microsoft.com/en-us/windows/win32/adschema/a-thumbnailphoto)].

    __uid__: String or null [[uid](https://docs.microsoft.com/en-us/windows/win32/adschema/a-uid)].

    __uidNumber__: String or null [[uidNumber](https://docs.microsoft.com/en-us/windows/win32/adschema/a-uidNumber)].

    __unixHomeDirectory__: String or null [[unixHomeDirectory](https://docs.microsoft.com/en-gb/windows/win32/adschema/a-unixHomeDirectory)].

    __userAccountControl__: Integer [[User-Account-Control](https://docs.microsoft.com/en-us/windows/win32/adschema/a-useraccountcontrol)].

    __userPrincipalName__: String or null [[User-Principal-Name](https://docs.microsoft.com/en-us/windows/win32/adschema/a-userprincipalname)].

??? info "Extra User Keys"

    The following keys are created by PPA based on other Active Directory attributes:

    __is_disabled__: Boolean indicating whether the user is disabled.

    __is_enabled__: Boolean indicating whether the user is enabled.

    __is_expired__: Boolean indicating whether the user is expired.

    __is_locked__: Boolean indicating whether the user is locked out.

    __password_age__: A dictionary containing the user's password age in `days`, `hours`, & `minutes`.


!!! tip "See more about Users [here](https://docs.microsoft.com/en-gb/windows/win32/adschema/c-user)"

    """

    accountExpires: str
    badPasswordTime: str
    badPwdCount: Optional[int]
    carLicense: Optional[List[str]]
    cn: Optional[str]
    countryCode: Optional[str]
    displayName: Optional[str]
    distinguishedName: str
    gidNumber: Optional[int]
    givenName: Optional[str]
    info: Optional[str]
    is_disabled: bool
    is_enabled: bool
    is_expired: bool
    is_locked: bool
    lastLogoff: str
    lastLogon: str
    lastLogonTimestamp: str
    lockoutTime: str
    loginShell: Optional[str]
    logonCount: Optional[int]
    mail: Optional[str]
    manager: Optional[str]
    memberOf: Optional[List[str]]
    name: Optional[str]
    objectGUID: str
    objectSid: str
    password_age: PasswordAge
    physicalDeliveryOfficeName: Optional[str]
    pwdLastSet: str
    sAMAccountName: str
    sAMAccountType: Optional[int]
    sn: Optional[str]
    telephoneNumber: Optional[str]
    thumbnailPhoto: Optional[str]
    uid: Optional[str]
    uidNumber: Optional[int]
    unixHomeDirectory: Optional[str]
    userAccountControl: int
    userPrincipalName: Optional[str]


class UserData(CastObject):
    __annotations__ = User.__dict__["__annotations__"]
    __class_config__ = {
        "cast_functions": {
            "fields": {
                "objectGUID": _get_decoded_ldap_string,
                "objectSid": _get_decoded_ldap_string,
                "thumbnailPhoto": _get_decoded_ldap_string,
                "accountExpires": _format_date_string,
                "badPasswordTime": _format_date_string,
                "lastLogon": _format_date_string,
                "lastLogoff": _format_date_string,
                "lastLogonTimestamp": _format_date_string,
                "lockoutTime": _format_date_string,
                "pwdLastSet": _format_date_string,
            },
        },
        "always_cast": ["objectGUID", "objectSid", "thumbnailPhoto"],
    }


class Group(TypedDict):
    """
- Contains Active Directory Group LDAP attributes

- Expand below to see each key and a link to the relevant Microsoft article

??? info "Active Directory Group Keys"

    __cn__: String or null [[Common-Name](https://docs.microsoft.com/en-us/windows/win32/adschema/a-cn)].

    __description__: String or null [[Description](https://docs.microsoft.com/en-us/windows/win32/adschema/a-description)].

    __distinguishedName__: String [[Obj-Dist-Name](https://docs.microsoft.com/en-us/windows/win32/adschema/a-distinguishedname)].

    __gidNumber__: Integer or null [[gidNumber](https://docs.microsoft.com/en-us/windows/win32/adschema/a-gidNumber)].

    __groupType__: String [[Group-Type](https://docs.microsoft.com/en-gb/windows/win32/adschema/a-grouptype)].

    __info__: String or null [[Comment](https://docs.microsoft.com/en-us/windows/win32/adschema/a-info)].

    __managedBy__: String or null [[ManagedBy](https://docs.microsoft.com/en-us/windows/win32/adschema/a-managedby)].

    __member__: List of [User](#user), [Group](#group), or [Computer](#computer) distinguished names or null [[Member](https://docs.microsoft.com/en-us/windows/win32/adschema/a-member)].

    __memberUid__: List of strings or null [[memberUid](https://docs.microsoft.com/en-us/windows/win32/adschema/a-memberuid)].

    __name__: String or null [[RDN](https://docs.microsoft.com/en-us/windows/win32/adschema/a-name)].

    __objectGUID__: String [[Object-Guid](https://docs.microsoft.com/en-us/windows/win32/adschema/a-objectguid)].

    __objectSid__: String [[Object-Sid](https://docs.microsoft.com/en-us/windows/win32/adschema/a-objectsid)].

    __sAMAccountName__: String [[Sam-Account-Name](https://docs.microsoft.com/en-us/windows/win32/adschema/a-samaccountname)].

    __sAMAccountType__: Integer [[Sam-Account-Type](https://docs.microsoft.com/en-us/windows/win32/adschema/a-samaccounttype)].

!!! warning "Group Membership Limit"

    Due to an Active Directory limitation, the __member__ key can only hold up to 1500 group members.

    If the group has more than 1500 members, only members 1-1500 will be included in this key.

    You can bypass this limitation by getting _nested_ users with the [get_members](../active_directory.groups/#get_members) action.


!!! tip "See more about Groups [here](https://docs.microsoft.com/en-gb/windows/win32/adschema/c-group)"

    """

    cn: Optional[str]
    description: Optional[str]
    distinguishedName: str
    gidNumber: Optional[int]
    groupType: str
    info: Optional[str]
    managedBy: Optional[str]
    member: Optional[List[str]]
    memberUid: Optional[List[str]]
    name: Optional[str]
    objectGUID: str
    objectSid: str
    sAMAccountName: str
    sAMAccountType: int


class GroupData(CastDataClass):
    __annotations__ = Group.__dict__["__annotations__"]
    __class_config__ = {
        "cast_functions": {"fields": {"objectGUID": _get_decoded_ldap_string, "objectSid": _get_decoded_ldap_string}},
        "always_cast": ["objectGUID", "objectSid"],
    }


class OrganizationalUnit(TypedDict):
    """
- Contains Active Directory Organizational Unit LDAP attributes

- Expand below to see each key and a link to the relevant Microsoft article

??? info "Active Directory Organizational Unit Keys"

    __distinguishedName__: String [[Obj-Dist-Name](https://docs.microsoft.com/en-us/windows/win32/adschema/a-distinguishedname)].

    __name__: String [[Organizational-Unit-Name](https://docs.microsoft.com/en-us/windows/win32/adschema/a-ou)].

    __objectGUID__: String [[Object-Guid](https://docs.microsoft.com/en-us/windows/win32/adschema/a-objectguid)].

    __ou__: String [[Organizational-Unit-Name](https://docs.microsoft.com/en-us/windows/win32/adschema/a-ou)].

!!! tip "See more about Organizational Units [here](https://docs.microsoft.com/en-gb/windows/win32/adschema/c-organizationalunit)"

    """

    distinguishedName: str
    name: Optional[str]
    objectGUID: str
    ou: Optional[str]


class OrganizationalUnitData(CastObject):
    __annotations__ = OrganizationalUnit.__dict__["__annotations__"]
    __class_config__ = {
        "cast_functions": {"fields": {"objectGUID": _get_decoded_ldap_string}},
        "always_cast": ["objectGUID"],
    }


class Computer(TypedDict):
    """
- Contains Active Directory Computer LDAP attributes

- Expand below to see each key and a link to the relevant Microsoft article

??? info "Active Directory Computer Keys"

    __accountExpires__: Date string or null [[Account-Expires](https://docs.microsoft.com/en-us/windows/win32/adschema/a-accountexpires)].

    __carLicense__: String or null [[carLicense](https://docs.microsoft.com/en-us/windows/win32/adschema/a-carlicense)].

    __cn__: String or null [[Common-Name](https://docs.microsoft.com/en-us/windows/win32/adschema/a-cn)].

    __description__: String or null [[Description](https://docs.microsoft.com/en-us/windows/win32/adschema/a-description)].

    __displayName__: String or null [[Display-Name](https://docs.microsoft.com/en-us/windows/win32/adschema/a-displayname)].

    __distinguishedName__: String [[Obj-Dist-Name](https://docs.microsoft.com/en-us/windows/win32/adschema/a-distinguishedname)].

    __info__: String or null [[Comment](https://docs.microsoft.com/en-us/windows/win32/adschema/a-info)].

    __lastLogoff__: Date string or null [[Last-Logoff](https://docs.microsoft.com/en-us/windows/win32/adschema/a-lastlogoff)].

    __lastLogon__: Date string or null [[Last-Logon](https://docs.microsoft.com/en-us/windows/win32/adschema/a-lastlogon)].

    __lastLogonTimestamp__: Date string or null [[Last-Logon-Timestamp](https://docs.microsoft.com/en-us/windows/win32/adschema/a-lastlogontimestamp)].

    __lockoutTime__: Date string or null [[Lockout-Time](https://docs.microsoft.com/en-us/windows/win32/adschema/a-lockouttime)].

    __logonCount__: Integer [[Logon-Count](https://docs.microsoft.com/en-us/windows/win32/adschema/a-logoncount)].

    __mail__: String or null [[E-mail-Addresses](https://docs.microsoft.com/en-us/windows/win32/adschema/a-mail)].

    __mail__: String or null [[Managed-By](https://docs.microsoft.com/en-us/windows/win32/adschema/a-managedby)].

    __memberOf__: List of [Group Dictionaries](#group) [[Is-Member-Of-DL](https://docs.microsoft.com/en-us/windows/win32/adschema/a-memberof)].

    __name__: String or null [[RDN](https://docs.microsoft.com/en-us/windows/win32/adschema/a-name)].

    __objectGUID__: String [[Object-Guid](https://docs.microsoft.com/en-us/windows/win32/adschema/a-objectguid)].

    __objectSid__: String [[Object-Sid](https://docs.microsoft.com/en-us/windows/win32/adschema/a-objectsid)].

    __operatingSystem__: String or null [[Operating-System](https://docs.microsoft.com/en-us/windows/win32/adschema/a-operatingsystem)].

    __operatingSystemHotfix__: String or null [[Operating-System-Hotfix](https://docs.microsoft.com/en-us/windows/win32/adschema/a-operatingsystemhotfix)].

    __operatingSystemServicePack__: String or null [[Operating-System-Service-Pack](https://docs.microsoft.com/en-us/windows/win32/adschema/a-operatingsystemservicepack)].

    __operatingSystemVersion__: String or null [[Operating-System-Version](https://docs.microsoft.com/en-us/windows/win32/adschema/a-operatingsystemversion)].

    __pwdLastSet__: Date string or null [[Pwd-Last-Set](https://docs.microsoft.com/en-us/windows/win32/adschema/a-pwdlastset)].

    __sAMAccountName__: String [[Sam-Account-Name](https://docs.microsoft.com/en-us/windows/win32/adschema/a-samaccountname)].

    __sAMAccountType__: Integer [[Sam-Account-Type](https://docs.microsoft.com/en-us/windows/win32/adschema/a-samaccounttype)].

    __userAccountControl__: Integer [[User-Account-Control](https://docs.microsoft.com/en-us/windows/win32/adschema/a-useraccountcontrol)].

    __userPrincipalName__: String or null [[User-Principal-Name](https://docs.microsoft.com/en-us/windows/win32/adschema/a-userprincipalname)].

!!! tip "See more about Computers [here](https://docs.microsoft.com/en-gb/windows/win32/adschema/c-computer)"
    """

    accountExpires: str
    carLicense: Optional[List[str]]
    cn: Optional[str]
    description: Optional[str]
    displayName: Optional[str]
    distinguishedName: str
    info: Optional[str]
    lastLogoff: Optional[str]
    lastLogon: str
    lastLogonTimestamp: str
    lockoutTime: str
    logonCount: Optional[int]
    mail: Optional[str]
    managedBy: Optional[str]
    memberOf: Optional[List[str]]
    name: Optional[str]
    objectGUID: str
    objectSid: str
    operatingSystem: Optional[str]
    operatingSystemHotfix: Optional[str]
    operatingSystemServicePack: Optional[str]
    operatingSystemVersion: Optional[str]
    pwdLastSet: str
    sAMAccountName: str
    sAMAccountType: int
    userAccountControl: int
    userPrincipalName: Optional[str]


class ComputerData(CastObject):
    __annotations__ = Computer.__dict__["__annotations__"]
    __class_config__ = {
        "cast_functions": {
            "fields": {
                "objectGUID": _get_decoded_ldap_string,
                "objectSid": _get_decoded_ldap_string,
                "accountExpires": _format_date_string,
                "lastLogon": _format_date_string,
                "lastLogoff": _format_date_string,
                "lastLogonTimestamp": _format_date_string,
                "lockoutTime": _format_date_string,
                "pwdLastSet": _format_date_string,
            },
        },
        "always_cast": ["objectGUID", "objectSid"],
    }
